{*********************************************************************
*
* Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved
*
* File: D3DFile.h
*
* Desc: Support code for loading DirectX .X files.
*
*********************************************************************
* timeline :
*   01/01/21 -> test -> mesh class successfully loaded and rendered
*   12/31/2k -> alpha test ... without system crash :)
*   12/29/2k -> rewritten from c++
*
* by orthkon * orthkon@mail.com * http://mp3.com/orthkon
* changes:  1. added path of mesh file to texture name in Create1
*           2. changed pchar to string
*           3. Delphi like variable names
*   J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de)
*
*********************************************************************}
unit D3DFile;

interface

uses Windows, sysUtils, Direct3D8, D3DX8, DXFile, Direct3DRM;

type
  TEnumMeshes = function (HInstance: Integer; Data: Pointer): Boolean;
  TEnumMeshesLW = function (HInstance: LongWord; Data: Pointer): Boolean;

  //-----------------------------------------------------------------------------
  // Name: class CD3DMesh
  // Desc: Class for loading and rendering file-based meshes
  //-----------------------------------------------------------------------------
  TD3DMesh = class
  private
    FMediapath : string;
  public
    strName : string;
    pSysMemMesh : ID3DXMESH;    // SysMem mesh, lives through resize
    pLocalMesh : ID3DXMESH;     // Local mesh, rebuilt on resize
    dwNumMaterials : Cardinal; // Materials for the mesh
    pMaterials : array of TD3DMATERIAL8;
    pTextures : array of IDIRECT3DTEXTURE8;
    bUseMaterials : Boolean;

    // Rendering
    function Render( d3dDevice : IDIRECT3DDEVICE8; bDrawOpaqueSubsets, bDrawAlphaSubsets : Boolean ) : HRESULT;

    // Mesh access
    function GetSysMemMesh : ID3DXMESH;
    function GetLocalMesh : ID3DXMESH;

    // Rendering options
    procedure UseMeshMaterials( bFlag : Boolean );
    function SetFVF(d3dDevice : IDIRECT3DDEVICE8; dwFVF : DWORD ) : HRESULT;

    // Initializing
    function RestoreDeviceObjects( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
    function InvalidateDeviceObjects : HRESULT;

    // Creation/destruction
    function Create1( d3dDevice : IDIRECT3DDEVICE8; strFilename : string) : HRESULT;
    function Create2( d3dDevice : IDIRECT3DDEVICE8; FileData : IDIRECTXFILEDATA  ) : HRESULT;
    function Destroy_ : HRESULT;

    constructor D3DMesh(strName,Mediapath : string);
    constructor Create (Mediapath : string); overload;
    destructor Destroy; override;
  end;

  TD3DFrame = class
  private
  public
    strName : string;
    mat : TD3DXMATRIX;
    pMesh : TD3DMesh;
//    strMeshName : PChar;
    pNext, pChild : TD3DFrame;

  	// Matrix access
    procedure SetMatrix( mat : TD3DXMATRIX  );
    function GetMatrix : TD3DXMATRIX;

    function FindMesh( strMeshName : PChar ) : TD3DMesh;
    function FindFrame( strFrameName : PChar ) : TD3DFrame;
    function EnumMeshes( func : TEnumMeshes; pContent : Pointer ) : Boolean; overload;
    function EnumMeshes( func : TEnumMeshesLW; pContent : Pointer ) : Boolean; overload;

    function Destroy_ : HRESULT;
    function RestoreDeviceObjects( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
    function InvalidateDeviceObjects : HRESULT;
    function Render( d3dDevice : IDIRECT3DDEVICE8; bDrawOpaqueSubsets, bDrawAlphaSubsets : Boolean ) : HRESULT;

    constructor Create( strName : PChar ); overload;
    destructor Destroy; override;
  end;

  TD3DFile = class( TD3DFrame )
    function LoadFrame( d3dDevice : IDIRECT3DDEVICE8; FileData : IDIRECTXFILEDATA; ParentFrame : TD3DFrame ) : HRESULT;
    function LoadMesh( d3dDevice : IDIRECT3DDEVICE8; FileData : IDIRECTXFILEDATA; ParentFrame : TD3DFrame ) : HRESULT;
  public
    function Create_( d3dDevice : IDIRECT3DDEVICE8; strFilename : PChar ) : HRESULT; overload;
    function Render( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
  end;

implementation

constructor TD3DMesh.Create (Mediapath : string);
begin
  strName := '';
  pSysMemMesh        := nil;
  pLocalMesh         := nil;
  dwNumMaterials     := 0;
  pTextures          := nil;
  bUseMaterials      := TRUE;
  FMediapath:=Mediapath;
end;

constructor TD3DMesh.D3DMesh(strName,Mediapath : string);
begin
  strName := pchar(strName);
  pSysMemMesh        := nil;
  pLocalMesh         := nil;
  dwNumMaterials     := 0;
  pTextures          := nil;
  bUseMaterials      := TRUE;
  FMediapath:=Mediapath;
end;

destructor TD3DMesh.Destroy;
begin
  Destroy_;
end;

function TD3DMesh.Create1( d3dDevice : IDIRECT3DDEVICE8; strFilename : string) : HRESULT;
var
  pMtrlBuffer, pt1 : ID3DXBUFFER;
  d3dxMtrl : PD3DXMATERIAL;
  strTexture : string;
  i : DWORD;
begin
  pMtrlBuffer := nil;
  // Load the mesh into system memory
  Result := D3DXLoadMeshFromX
    (pchar(FMediapath+strFilename), D3DXMESH_SYSTEMMEM, d3dDevice, pt1, pMtrlBuffer,
     dwNumMaterials, pSysMemMesh );
  if Failed( Result ) then Exit;
  // Get material info for the mesh
  // Get the array of materials out of the buffer
  if ( pMtrlBuffer <> nil ) and ( dwNumMaterials > 0 ) then begin
    // Allocate memory for the materials and textures
    d3dxMtrl := pMtrlBuffer.GetBufferPointer;
    Result := E_FAIL;
    try
      SetLength( pMaterials, dwNumMaterials );
      SetLength( pTextures, dwNumMaterials );
    except
      Exit;
    end;
    // Copy each material and create it's texture
    for i := 0 to dwNumMaterials - 1 do begin
      // Copy the material
      pMaterials[i]         := d3dxMtrl.MatD3D;
      pMaterials[i].Ambient := pMaterials[i].Diffuse;
      pTextures[i]          := nil;
      // Create a texture
      if d3dxMtrl^.pTextureFilename <> nil then begin
      // load texture from same path as mesh
        strTexture := FMediapath+d3dxMtrl^.pTextureFilename;
        if Failed( D3DXCreateTextureFromFile( d3dDevice, pchar(strTexture), pTextures[i] ) ) then
          pTextures[i] := nil;
      end;
      Inc( d3dxMtrl );
    end;
  end;
  pMtrlBuffer := nil;
  Result := S_OK;
end;

function TD3DMesh.Create2( d3dDevice : IDIRECT3DDEVICE8; FileData : IDIRECTXFILEDATA  ) : HRESULT;
var
  pMtrlBuffer, pt1 : ID3DXBUFFER;
  d3dxMtrl : PD3DXMATERIAL;
  strTexture : PChar;
  i : DWORD;
begin
  pMtrlBuffer := nil;
  // Load the mesh from the DXFILEDATA object
  Result := D3DXLoadMeshFromXof( FileData, D3DXMESH_SYSTEMMEM, d3dDevice,
                              pt1, pMtrlBuffer, dwNumMaterials,
                              pSysMemMesh );
  if Failed( Result ) then Exit;
  // Get material info for the mesh
  // Get the array of materials out of the buffer
  if ( pMtrlBuffer <> nil ) and ( dwNumMaterials > 0 ) then begin
    // Allocate memory for the materials and textures
    d3dxMtrl := pMtrlBuffer.GetBufferPointer;
    Result := E_FAIL;
    try
      SetLength( pMaterials, dwNumMaterials );
      SetLength( pTextures, dwNumMaterials );
    except
      Exit;
    end;
    // Copy each material and create it's texture
    for i := 0 to dwNumMaterials - 1 do begin
      // Copy the material
      pMaterials[i]         := d3dxMtrl^.MatD3D;
      pMaterials[i].Ambient := pMaterials[i].Diffuse;
      pTextures[i]          := nil;
      // Create a texture
      if d3dxMtrl^.pTextureFilename <> nil then begin
        strTexture := d3dxMtrl^.pTextureFilename;
        if Failed( D3DXCreateTextureFromFile( d3dDevice, strTexture, pTextures[i] ) ) then
          pTextures[i] := nil;
      end;
      Inc( d3dxMtrl );
    end;
  end;
  pMtrlBuffer := nil;
  Result := S_OK;
end;

function TD3DMesh.SetFVF( d3dDevice : IDIRECT3DDEVICE8; dwFVF : DWORD ) : HRESULT;
var
  pTempSysMemMesh, pTempLocalMesh : ID3DXMESH;
begin
  Result := E_FAIL;
  pTempSysMemMesh := nil;
  pTempLocalMesh := nil;
  if pSysMemMesh <> nil then
    if Failed( pSysMemMesh.CloneMeshFVF( D3DXMESH_SYSTEMMEM, dwFVF, d3dDevice, pTempSysMemMesh ) ) then
      Exit;
    if pLocalMesh <> nil then
      if Failed( pLocalMesh.CloneMeshFVF( 0, dwFVF, d3dDevice, pTempLocalMesh ) ) then begin
        pTempSysMemMesh := nil;
        Exit;
      end;
  pSysMemMesh := nil;
  pLocalMesh := nil;
  if pTempSysMemMesh <> nil then pSysMemMesh := pTempSysMemMesh;
  if pTempLocalMesh <> nil then pLocalMesh := pTempLocalMesh;
  // Compute normals in case the meshes have them
  if pSysMemMesh <> nil then D3DXComputeNormals( ID3DXBASEMESH( pSysMemMesh ) );
  if pLocalMesh <> nil then D3DXComputeNormals( ID3DXBASEMESH( pLocalMesh ) );
  Result := S_OK;
end;

function TD3DMesh.RestoreDeviceObjects( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
begin
  Result := E_FAIL;
  if nil = pSysMemMesh then Exit;
  // Make a local memory version of the mesh. Note: because we are passing in
  // no flags, the default behavior is to clone into local memory.
  if Failed( pSysMemMesh.CloneMeshFVF
    ( 0, pSysMemMesh.GetFVF(), d3dDevice, pLocalMesh ) ) then Exit;
  Result := S_OK;
end;

function TD3DMesh.InvalidateDeviceObjects : HRESULT;
begin
  pLocalMesh := nil;
  Result := S_OK;
end;

function TD3DMesh.Destroy_ : HRESULT;
var
  i : Word;
begin
  InvalidateDeviceObjects;
  i := 0;
  while i < dwNumMaterials do begin
    pTextures[i] := nil;
    inc( i );
  end;
  pTextures := nil;
  pMaterials := nil;
  pSysMemMesh := nil;
  dwNumMaterials := 0;
  Result := S_OK;
end;

function TD3DMesh.Render( d3dDevice : IDIRECT3DDEVICE8; bDrawOpaqueSubsets, bDrawAlphaSubsets : Boolean ) : HRESULT;
var
  i : Cardinal;
begin
  Result := E_FAIL;
  if nil = pLocalMesh then Exit;
  // First, draw the subsets without alpha
  if bDrawOpaqueSubsets then begin
    for i := 0 to dwNumMaterials - 1 do begin
      if bUseMaterials then begin
        if pMaterials[i].Diffuse.a < 1.0 then continue;
        d3dDevice.SetMaterial( pMaterials[i] );
        d3dDevice.SetTexture( 0, pTextures[i] );
      end;
      pLocalMesh.DrawSubset( i );
    end;
  end;
  // Then, draw the subsets with alpha
  if bDrawAlphaSubsets and bUseMaterials then begin
    for i := 0 to dwNumMaterials - 1 do begin
      if pMaterials[i].Diffuse.a = 1.0 then continue;
      // Set the material and texture
      d3dDevice.SetMaterial( pMaterials[i] );
      d3dDevice.SetTexture( 0, pTextures[i] );
      pLocalMesh.DrawSubset( i );
    end;
  end;
  Result := S_OK;
end;

function TD3DMesh.GetSysMemMesh : ID3DXMESH;
begin
  Result := pSysMemMesh;
end;

function TD3DMesh.GetLocalMesh : ID3DXMESH;
begin
  Result := pLocalMesh;
end;

procedure TD3DMesh.UseMeshMaterials( bFlag : Boolean );
begin
  bUseMaterials := bFlag;
end;

constructor TD3DFrame.Create( strName : PChar );
begin
  strName := strName;
  D3DXMatrixIdentity( mat );
  pMesh := nil;
  pChild := nil;
  pNext  := nil;
end;

destructor TD3DFrame.Destroy;
begin
  pChild := nil;
  pNext := nil;
end;

// is that right??? no :<
function TD3DFrame.EnumMeshes( func : TEnumMeshes; pContent : Pointer ) : Boolean;
begin
  Result := EnumMeshes( TEnumMeshesLW( func ), pContent );
  //if pMesh <> nil then EnumMeshCB( pMesh, pContext );
  if pChild <> nil then pChild.EnumMeshes( TEnumMeshes( func ), pContext( pContent ) );
  if pNext <> nil then pNext.EnumMeshes( TEnumMeshes( func ), pContext( pContent ) );
end;

function TD3DFrame.EnumMeshes( func : TEnumMeshesLW; pContent : Pointer ) : Boolean;
begin
  Result := TRUE;
end;

function TD3DFrame.FindMesh( strMeshName : PChar ) : TD3DMesh;
begin
  Result := nil;
  if pMesh <> nil then
    if pMesh.strName = strMeshName then begin
      Result := pMesh;
      Exit;
    end;
  if pChild <> nil then begin
    Result := pChild.FindMesh( strMeshName );
    if Result <> nil Then Exit;
  end;
  if pNext <> nil then begin
    Result := pNext.FindMesh( strMeshName );
    if Result <> nil Then Exit;
  end;
end;

function TD3DFrame.FindFrame( strFrameName : PChar ) : TD3DFrame;
begin
  Result := nil;
  if strName = strFrameName then begin
    Result := Self;
    Exit;
  end;
  if pChild <> nil then begin
    Result := pChild.FindFrame( strFrameName );
    if Result <> nil then Exit;
  end;
  if pNext <> nil then begin
    Result := pNext.FindFrame( strFrameName );
    if Result <> nil then Exit;
  end;
end;

function TD3DFrame.Destroy_ : HRESULT;
begin
  if pMesh <> nil then pMesh.Destroy_;
  if pChild <> nil then pChild.Destroy_;
  if pNext <> nil then pNext.Destroy_;
  pMesh := nil;
  pChild := nil;
  pNext := nil;
  Result := S_OK;
end;

function TD3DFrame.RestoreDeviceObjects( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
begin
  if pMesh <> nil then pMesh.RestoreDeviceObjects( d3dDevice );
  if pChild <> nil then pChild.RestoreDeviceObjects( d3dDevice );
  if pNext <> nil then pNext.RestoreDeviceObjects( d3dDevice );
  Result := S_OK;
end;

function TD3DFrame.InvalidateDeviceObjects : HRESULT;
begin
  if pMesh <> nil then pMesh.InvalidateDeviceObjects;
  if pChild <> nil then pChild.InvalidateDeviceObjects;
  if pNext <> nil then pNext.InvalidateDeviceObjects;
  Result := S_OK;
end;

function TD3DFrame.Render( d3dDevice : IDIRECT3DDEVICE8; bDrawOpaqueSubsets, bDrawAlphaSubsets : Boolean ) : HRESULT;
var
  matSavedWorld, matWorld : TD3DXMATRIX;
begin
  d3dDevice.GetTransform( D3DTS_WORLD, matSavedWorld );
  D3DXMatrixMultiply( matWorld, mat, matSavedWorld );
  d3dDevice.SetTransform( D3DTS_WORLD, matWorld );
  if pMesh <> nil then
    pMesh.Render( d3dDevice, bDrawOpaqueSubsets, bDrawAlphaSubsets );
  if pChild <> nil then
    pChild.Render( d3dDevice, bDrawOpaqueSubsets, bDrawAlphaSubsets );
  d3dDevice.SetTransform( D3DTS_WORLD, matSavedWorld );
  if pNext <> nil then
    pNext.Render( d3dDevice, bDrawOpaqueSubsets, bDrawAlphaSubsets );
  Result := S_OK;
end;

procedure TD3DFrame.SetMatrix( mat : TD3DXMATRIX  );
begin
  mat := mat;
end;

function TD3DFrame.GetMatrix : TD3DXMATRIX;
begin
  Result := mat;
end;

function CheckGUID( p1, p2 : TGUID ) : Boolean;
var
  i : Byte;
begin
  Result := False;
  for i := 0 to 7 do if p1.D4[i] <> p2.D4[i] then Exit;
  Result := ( p1.D1 = p2.D1 ) and ( p1.D2 = p2.D2 ) and ( p1.D3 = p2.D3 );
end;

function TD3DFile.LoadFrame( d3dDevice : IDIRECT3DDEVICE8;
                             FileData : IDIRECTXFILEDATA;
                             ParentFrame : TD3DFrame ) : HRESULT;
var
  ChildData : IDIRECTXFILEDATA;
  ChildObj : IDIRECTXFILEOBJECT;
  ppGUID : PGUID;
  cbSize : Cardinal;
  CurrentFrame : TD3DFrame;
  matMatrix : PD3DXMATRIX;
  dwNameLength : PDWORD;
  strName : array [0..255] of char;
begin
  // Get the type of the object
  Result := FileData.GetType( ppGUID );
  if Failed( Result ) then Exit;
  if checkGUID( ppGUID^, TID_D3DRMMesh ) then begin
    Result := LoadMesh( d3dDevice, FileData, ParentFrame );
    if Failed( Result ) then Exit;
  end;
  if checkGUID( ppGUID^, TID_D3DRMFrameTransformMatrix ) then begin
    Result := FileData.GetData( nil, cbSize, Pointer( matMatrix ) );
    if Failed( Result ) then Exit;
    // Update the parent's matrix with the new one
    ParentFrame.SetMatrix( matMatrix^ );
  end;
  if checkGUID( ppGUID^, TID_D3DRMFrame ) then begin
    // Get the frame name
    dwNameLength := nil;
    FileData.GetName( nil, dwNameLength );
    if dwNameLength^ > 0 then
      FileData.GetName( strName, dwNameLength );
    // Create the frame
    CurrentFrame := TD3DFrame.Create( strName );
    CurrentFrame.pNext := ParentFrame.pChild;
    ParentFrame.pChild := CurrentFrame;
    // Enumerate child objects
    while( SUCCEEDED( FileData.GetNextObject( ChildObj ) ) ) do begin
      // Query the child for its FileData
      Result := ChildObj.QueryInterface( IID_IDirectXFileData, ChildData );
      if SUCCEEDED( Result ) then begin
        Result := LoadFrame( d3dDevice, ChildData, CurrentFrame );
        ChildData := nil;
      end;
      ChildObj := nil;
      if FAILED( Result ) then Exit;
    end;
  end;
  Result := S_OK;
end;

function TD3DFile.LoadMesh( d3dDevice : IDIRECT3DDEVICE8;
                            FileData : IDIRECTXFILEDATA;
                            ParentFrame : TD3DFrame ) : HRESULT;
var
  dwNameLength : PDWORD;
  strName : array [0..255] of char;
begin
  Result := E_FAIL;
  // Currently only allowing one mesh per frame
  if ParentFrame.pMesh <> nil then Exit;
  // Get the mesh name
  dwNameLength := nil;
  FileData.GetName( nil, dwNameLength );
  if dwNameLength^ > 0 then
    FileData.GetName( strName, dwNameLength );
  // Create the mesh
  ParentFrame.pMesh := TD3DMesh.D3DMesh( strName,'');
  ParentFrame.pMesh.Create2( d3dDevice, FileData );
  Result := S_OK;
end;

function TD3DFile.Create_( d3dDevice : IDIRECT3DDEVICE8; strFilename : PChar ) : HRESULT;
var
  DXFile : IDIRECTXFILE;
  EnumObj : IDIRECTXFILEENUMOBJECT;
  FileData : IDIRECTXFILEDATA;
begin
  // Create a x file object
  Result := DirectXFileCreate( DXFile );
  if Failed( Result ) then Exit;
  // Register templates for d3drm and patch extensions.
  Result := DXFile.RegisterTemplates( @D3DRM_XTEMPLATES, D3DRM_XTEMPLATE_BYTES );
  if Failed( Result ) then Begin
    DXFile := nil;
    Exit;
  end;
  // Create enum object
  Result := DXFile.CreateEnumObject( @strFilename, DXFILELOAD_FROMFILE, EnumObj );
  if Failed( Result ) then Begin
    DXFile := nil;
    Exit;
  end;
  // Enumerate top level objects (which are always frames)
  while( SUCCEEDED( EnumObj.GetNextDataObject( FileData ) ) ) do begin
    Result := LoadFrame( d3dDevice, FileData, Self );
    FileData := nil;
    if Failed( Result ) then Begin
      EnumObj := nil;
      DXFile := nil;
      Exit;
    end;
  end;
  FileData := nil;
  EnumObj := nil;
  DXFile := nil;
  Result := S_OK;
end;

function TD3DFile.Render( d3dDevice : IDIRECT3DDEVICE8 ) : HRESULT;
var
  matSavedWorld, matWorld : TD3DXMATRIX;
begin
  // Set up the world transformation
  d3dDevice.GetTransform( D3DTS_WORLD, matSavedWorld );
  D3DXMatrixMultiply( matWorld, matSavedWorld, mat );
  d3dDevice.SetTransform( D3DTS_WORLD, matWorld );
  // Render opaque subsets in the meshes
  if pChild <> nil then
      pChild.Render( d3dDevice, TRUE, FALSE );
  // Enable alpha blending
  d3dDevice.SetRenderState( D3DRS_ALPHABLENDENABLE, 1 );
  d3dDevice.SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
  d3dDevice.SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
  // Render alpha subsets in the meshes
  if pChild <> nil then pChild.Render( d3dDevice, FALSE, TRUE );
  // Restore state
  d3dDevice.SetRenderState( D3DRS_ALPHABLENDENABLE, 0 );
  d3dDevice.SetTransform( D3DTS_WORLD, matSavedWorld );
  Result := S_OK;
end;

end.
